//
// Ray Tracer by SSJX (https://ssjx.co.uk) 
//
// Based on the Business Card Ray Tracer created by Andrew Kensler and unobfuscated by Fabien Sanglard.
//
// For more information:
// 		https://fabiensanglard.net/rayTracing_back_of_business_card/
//
// Fabien Sanglard's C++ version was used as the reference version, some of the comments here are from that.
// 
// Thank you both!
//
// Compiles with:
// 		dmd -i trace.d
//


import std.stdio;
import std.random;
import std.math;
import std.algorithm;	// min
import std.parallelism;
import std.datetime.stopwatch;

import bmp;

//Define a vector class
struct Vec{
    float x, y, z;

   this(float a, float b, float c) {
      x = a;
      y = b;
      z = c;
    }
	
	// Vector add
	Vec opBinary(string s)(Vec r) if (s == "+"){
		return Vec(x + r.x, y + r.y, z + r.z);
    }

	// Vector dot product
  	float opBinary(string s)(Vec r) if (s == "%"){
		return (x * r.x) + (y * r.y) + (z * r.z);
    }
  
	// Vector scaling
	Vec opBinary(string s)(float r) if (s == "*"){
		return Vec(x * r, y * r, z * r);
    }

	// Cross-product
	Vec opBinary(string s)(Vec r) if (s == "^"){
		return Vec(y*r.z-z*r.y, z*r.x-x*r.z, x*r.y-y*r.x);
    }
}

// Inverse Square Root
Vec inverse(Vec r){
	float sq=1 / sqrt( r%r);
	return r*sq;
}

//The set of sphere positions describing the world.
//Those integers are in fact bit vectors.

/*
int[] G=[247570,280596,280600,249748,18578,18577,231184,16,16];

16                    1    
16                    1    
231184   111    111   1    
18577       1  1   1  1   1
18578       1  1   1  1  1 
249748   1111  11111  1 1  
280600  1   1  1      11   
280596  1   1  1      1 1  
247570   1111   111   1  1 

*/

static int[9] G=[143132,141474,141472,257824,141472,141474,118556,0,0];
/*
0011100111100011100 = 118556
0100010100010100010 = 141474
0100010100010100000 = 141472
0111110111100100000 = 257824
0100010100010100000 = 141472
0100010100010100010 = 141474
0100010111100011100 = 143132
*/



// Random generator, return a float within range [0-1]
float randomVal() { return uniform(0.0f, 1.0f); }

void main(){
	const int w = 512;
	const int h = 512;
	const int rays = 64;
	ubyte[] output=new ubyte[w*h*3];
		
	Vec g=inverse(Vec(-6,-16,0));       // Camera direction
	Vec a=inverse(Vec(0,0,1)^g)*.002; 	// Camera up vector...Seem Z is pointing up :/ WTF !
	Vec b=inverse(g^a)*.002;        	// The right vector, obtained via traditional cross-product
	Vec c=(a+b)*-256+g;       			// WTF ? See https://news.ycombinator.com/item?id=6425965.

	writeln("Ray Tracer v0.1 (16/02/2024) by ssjx (https://ssjx.co.uk)");
	writeln("=========================================================");
	writeln("Built with the ",__VENDOR__," Compiler");
	writeln("Please wait, this may take some time...");
	
	StopWatch sw;
	sw.start();

	// === PARALLEL ===
	struct mypoint{
		int x;
		int y;
		int pos;
	}
	
	mypoint[] points=new mypoint[w*h];
	
	auto n=0;
	foreach_reverse(y;0..h){
		auto px=y*(w*3);
		foreach_reverse(x;0..w){
			points[n].x=x;
			points[n].y=y;
			points[n].pos=px;
			px+=3;
			n++;
		}
	}

	
	foreach(pt;parallel(points)){
	   	auto x=pt.x;
		auto y=pt.y;
		auto px=pt.pos;
		
		//Reuse the vector class to store not XYZ but a RGB pixel color
		Vec p=Vec(13,13,13);     // Default pixel color is almost pitch black
	   
		//Cast 64 rays per pixel (For blur (stochastic sampling) and soft-shadows. 
		foreach_reverse(r;0..rays){ 
		  
			// The delta to apply to the origin of the view (For Depth of View blur).
			Vec t=a*(randomVal()-.5)*99+b*(randomVal()-.5)*99; // A little bit of delta 
										   
			// Set the camera focal point v(17,16,8) and Cast the ray 
			// Accumulate the color returned in the p variable
			p=Sample(Vec(17,16,8)+t, //Ray Origin
				inverse(t*-1+(a*(randomVal()+x)+b*(y+randomVal())+c)*16) 	// Ray Direction with random deltas
																			// for stochastic sampling
				)*3.5+p; // +p for color accumulation
		}
	   
		// Write our bitmap
		output[px]=cast(ubyte)p.z;
		output[px+1]=cast(ubyte)p.y;
		output[px+2]=cast(ubyte)p.x;
	}


	// === SINGLE ===
	// foreach_reverse(y;0..h){    //For each column
		// auto px=y*(w*3);

		// foreach_reverse(x;0..w){   //For each pixel in a line

			// //Reuse the vector class to store not XYZ but a RGB pixel color
			// Vec p=Vec(13,13,13);     // Default pixel color is almost pitch black

			// //Cast 64 rays per pixel (For blur (stochastic sampling) and soft-shadows. 
			// foreach_reverse(r;0..rays){ 
				// // The delta to apply to the origin of the view (For Depth of View blur).
				// Vec t=a*(randomVal()-.5)*99+b*(randomVal()-.5)*99; // A little bit of delta 

				// // Set the camera focal point v(17,16,8) and Cast the ray 
				// // Accumulate the color returned in the p variable
				// p=Sample(Vec(17,16,8)+t, //Ray Origin
				// inverse(t*-1+(a*(randomVal()+x)+b*(y+randomVal())+c)*16) 	// Ray Direction with random deltas
																			// // for stochastic sampling
				// )*3.5+p; // +p for color accumulation
			// }

			// // Write our bitmap
			// output[px]=cast(ubyte)p.z;
			// output[px+1]=cast(ubyte)p.y;
			// output[px+2]=cast(ubyte)p.x;
			// px+=3;

		// }
	// }
		// =================
	
	long msecs = sw.peek.total!"msecs";
	writeln("Elapsed time: ",msecs,"ms");
	
	save_bmp(w,h,output,"output");
}


// (S)ample the world and return the pixel color for
// a ray passing by point o (Origin) and d (Direction)
Vec Sample(Vec o,Vec d){
	float t;
	Vec n;

	//Search for an intersection ray Vs World.
	int m=Test(o,d,t,n);

	if(!m) // m==0
	//No sphere found and the ray goes upward: Generate a sky color  
	return Vec(.7,.6,1)*pow(1-d.z,4);

	//A sphere was maybe hit.

	Vec h=o+d*t;                    // h = intersection coordinate
	Vec l=inverse(Vec(9+randomVal(),9+randomVal(),16)+h*-1);  // 'l' = direction to light (random delta for shadows).
	Vec r=d+n*(n%d*-2);               // r = The half-vector

	//Calculated the lambertian factor
	float b=l%n;

	//Calculate illumination factor (lambertian coefficient > 0 or in shadow)?
	if(b<0||Test(h,l,t,n)){
		b=0;
	}

	// Calculate the color 'p' with diffuse and specular component 
	float p=pow(l%r*(b>0),99);

	if(m&1){   //m == 1
		h=h*.2; //No sphere was hit and the ray was going downward: Generate a floor color
		return(cast(int)(ceil(h.x)+ceil(h.y))&1?Vec(3,1,1):Vec(3,3,3))*(b*.2+.1);
	}

	//m == 2 A sphere was hit. Cast an ray bouncing from the sphere surface.
	return Vec(p,p,p)+Sample(h,r)*.5; //Attenuate color by 50% since it is bouncing (* .5)
}
  
// The intersection test for line [o,v].
// Return 2 if a hit was found (and also return distance t and bouncing ray n).
// Return 0 if no hit was found but ray goes upward
// Return 1 if no hit was found but ray goes downward
int Test(Vec o,Vec d,ref float t,ref Vec n){ 
	t=float.max;
	int m=0;
	float pf=-o.z/d.z;
	if(.01<pf){
		t=pf;
		n=Vec(0,0,1);
		m=1;
	}
	
	//The world is encoded in G, with 9 lines and 19 columns
	foreach(k;0..19){  //For each columns of objects
		foreach(j;0..9){   //For each line on that columns

			if(G[j]&1<<k){ //For this line j, is there a sphere at column i ?

				// There is a sphere but does the ray hits it ?
				Vec p=o+Vec(-k,0,-j-4);
				float b=p%d;
				float c=p%p-1;
				float q=b*b-c;

				//Does the ray hit the sphere ?
				if(q>0){
					//It does, compute the distance camera-sphere
					float s=-b-sqrt(q);

					if(s<t && s>.01){
						// So far this is the minimum distance, save it. And also
						// compute the bouncing ray vector into 'n'  
						t=s;
						n=inverse(p+d*t);
						m=2;
					}
				}
			}
		}
	}
	return m;
}
